@@ -32,13 +32,12 @@ NK_API void                 nk_glfw3_device_destroy(void);
 NK_API void                 nk_glfw3_device_create(void);
 
 NK_API void                 nk_glfw3_char_callback(GLFWwindow *win, unsigned int codepoint);
+NK_API void                 nk_glfw3_key_callback(GLFWwindow *win, int key, int scancode, int action, int mods);
 NK_API void                 nk_gflw3_scroll_callback(GLFWwindow *win, double xoff, double yoff);
 NK_API void                 nk_glfw3_mouse_button_callback(GLFWwindow *win, int button, int action, int mods);
 
-NK_API GLuint               nk_glfw3_get_tex_ogl_id(int tex_index);
-NK_API GLuint64             nk_glfw3_get_tex_ogl_handle(int tex_index);
-NK_API int                  nk_glfw3_create_texture(const void* image, int width, int height);
-NK_API void                 nk_glfw3_destroy_texture(int tex_index);
+NK_API unsigned int                  nk_glfw3_create_texture(const void* image, int width, int height);
+NK_API void                 nk_glfw3_destroy_texture(unsigned int tex_index);
 
 #endif
 /*
@@ -50,6 +49,9 @@ NK_API void                 nk_glfw3_destroy_texture(int tex_index);
  */
 #ifdef NK_GLFW_GL4_IMPLEMENTATION
 #undef NK_GLFW_GL4_IMPLEMENTATION
+#include <stdio.h>
+#include <stdlib.h>
+#include <assert.h>
 
 #ifndef NK_GLFW_TEXT_MAX
 #define NK_GLFW_TEXT_MAX 256
@@ -60,9 +62,6 @@ NK_API void                 nk_glfw3_destroy_texture(int tex_index);
 #ifndef NK_GLFW_DOUBLE_CLICK_HI
 #define NK_GLFW_DOUBLE_CLICK_HI 0.2
 #endif
-#ifndef NK_GLFW_MAX_TEXTURES
-#define NK_GLFW_MAX_TEXTURES 256
-#endif
 
 struct nk_glfw_vertex {
     float position[2];
@@ -72,7 +71,7 @@ struct nk_glfw_vertex {
 
 struct nk_glfw_device {
     struct nk_buffer cmds;
-    struct nk_draw_null_texture null;
+    struct nk_draw_null_texture tex_null;
     GLuint vbo, vao, ebo;
     GLuint prog;
     GLuint vert_shdr;
@@ -88,8 +87,6 @@ struct nk_glfw_device {
     struct nk_glfw_vertex *vert_buffer;
     int *elem_buffer;
     GLsync buffer_sync;
-    GLuint tex_ids[NK_GLFW_MAX_TEXTURES];
-    GLuint64 tex_handles[NK_GLFW_MAX_TEXTURES];
 };
 
 static struct nk_glfw {
@@ -101,16 +98,17 @@ static struct nk_glfw {
     struct nk_font_atlas atlas;
     struct nk_vec2 fb_scale;
     unsigned int text[NK_GLFW_TEXT_MAX];
+    nk_char key_events[NK_KEY_MAX];
     int text_len;
     struct nk_vec2 scroll;
     double last_button_click;
     int is_double_click_down;
     struct nk_vec2 double_click_pos;
+    float delta_time_seconds_last;
 } glfw;
 
 #define NK_SHADER_VERSION "#version 450 core\n"
 #define NK_SHADER_BINDLESS "#extension GL_ARB_bindless_texture : require\n"
-#define NK_SHADER_64BIT "#extension GL_ARB_gpu_shader_int64 : require\n"
 
 NK_API void
 nk_glfw3_device_create()
@@ -120,7 +118,6 @@ nk_glfw3_device_create()
     static const GLchar *vertex_shader =
         NK_SHADER_VERSION
         NK_SHADER_BINDLESS
-        NK_SHADER_64BIT
         "uniform mat4 ProjMtx;\n"
         "in vec2 Position;\n"
         "in vec2 TexCoord;\n"
@@ -135,15 +132,13 @@ nk_glfw3_device_create()
     static const GLchar *fragment_shader =
         NK_SHADER_VERSION
         NK_SHADER_BINDLESS
-        NK_SHADER_64BIT
         "precision mediump float;\n"
-        "uniform uint64_t Texture;\n"
+        "layout(bindless_sampler) uniform sampler2D Texture;\n"
         "in vec2 Frag_UV;\n"
         "in vec4 Frag_Color;\n"
         "out vec4 Out_Color;\n"
         "void main(){\n"
-        "   sampler2D smp = sampler2D(Texture);\n"
-        "   Out_Color = Frag_Color * texture(smp, Frag_UV.st);\n"
+        "   Out_Color = Frag_Color * texture(Texture, Frag_UV.st);\n"
         "}\n";
 
     struct nk_glfw_device *dev = &glfw.ogl;
@@ -151,12 +146,10 @@ nk_glfw3_device_create()
     dev->prog = glCreateProgram();
     dev->vert_shdr = glCreateShader(GL_VERTEX_SHADER);
     dev->frag_shdr = glCreateShader(GL_FRAGMENT_SHADER);
-    glShaderSource(dev->vert_shdr, 1, &vertex_shader, 0);
-    glShaderSource(dev->frag_shdr, 1, &fragment_shader, 0);
-    glCompileShader(dev->vert_shdr);
-    glCompileShader(dev->frag_shdr);
-    glGetShaderiv(dev->vert_shdr, GL_COMPILE_STATUS, &status);
 
+    glShaderSource(dev->vert_shdr, 1, &vertex_shader, 0);
+    glCompileShader(dev->vert_shdr);
+    glGetShaderiv(dev->vert_shdr, GL_COMPILE_STATUS, &status);
     glGetShaderiv(dev->vert_shdr, GL_INFO_LOG_LENGTH, &len);
     if (len > 1) {
         char *log = (char*)calloc((size_t)len, sizeof(char));
@@ -164,7 +157,11 @@ nk_glfw3_device_create()
         fprintf(stdout, "[GL]: failed to compile shader: %s", log);
         free(log);
     }
+    assert(status == GL_TRUE);
 
+    glShaderSource(dev->frag_shdr, 1, &fragment_shader, 0);
+    glCompileShader(dev->frag_shdr);
+    glGetShaderiv(dev->frag_shdr, GL_COMPILE_STATUS, &status);
     glGetShaderiv(dev->frag_shdr, GL_INFO_LOG_LENGTH, &len);
     if (len > 1) {
         char *log = (char*)calloc((size_t)len, sizeof(char));
@@ -172,10 +169,8 @@ nk_glfw3_device_create()
         fprintf(stdout, "[GL]: failed to compile shader: %s", log);
         free(log);
     }
+    assert(status == GL_TRUE);
 
-    assert(status == GL_TRUE);
-    glGetShaderiv(dev->frag_shdr, GL_COMPILE_STATUS, &status);
-    assert(status == GL_TRUE);
     glAttachShader(dev->prog, dev->vert_shdr);
     glAttachShader(dev->prog, dev->frag_shdr);
     glLinkProgram(dev->prog);
@@ -228,51 +223,16 @@ nk_glfw3_device_create()
     dev->vert_buffer = (struct nk_glfw_vertex*) glMapNamedBufferRange(dev->vbo, 0, vb_size, flags);
     dev->elem_buffer = (int*) glMapNamedBufferRange(dev->ebo, 0, eb_size, flags);}
 
-    memset(dev->tex_ids, 0, sizeof(dev->tex_ids));
-    memset(dev->tex_handles, 0, sizeof(dev->tex_handles));
 }
 
-NK_INTERN int
-nk_glfw3_get_available_tex_index()
-{
-    int i = 0;
-    struct nk_glfw_device *dev = &glfw.ogl;
-    for (i = 0; i < NK_GLFW_MAX_TEXTURES; i++) {
-        if (dev->tex_ids[i] == 0)
-            return i;
-    } assert(0); /* max textures reached */
-    return -1;
-}
-
-NK_API GLuint
-nk_glfw3_get_tex_ogl_id(int tex_index)
-{
-    struct nk_glfw_device *dev = &glfw.ogl;
-    assert(tex_index >= 0 && tex_index < NK_GLFW_MAX_TEXTURES);
-    return dev->tex_ids[tex_index];
-}
-
-NK_API GLuint64
-nk_glfw3_get_tex_ogl_handle(int tex_index)
-{
-    struct nk_glfw_device *dev = &glfw.ogl;
-    assert(tex_index >= 0 && tex_index < NK_GLFW_MAX_TEXTURES);
-    return dev->tex_handles[tex_index];
-}
-
-NK_API int
+NK_API unsigned int
 nk_glfw3_create_texture(const void* image, int width, int height)
 {
     GLuint id;
     GLsizei w = (GLsizei)width;
     GLsizei h = (GLsizei)height;
-    struct nk_glfw_device *dev = &glfw.ogl;
-    int tex_index = nk_glfw3_get_available_tex_index();
-    if (tex_index < 0)
-        return -1;
 
     glCreateTextures(GL_TEXTURE_2D, 1, &id);
-    dev->tex_ids[tex_index] = id;
 
     glTextureParameteri(id, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
     glTextureParameteri(id, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
@@ -284,23 +244,15 @@ nk_glfw3_create_texture(const void* image, int width, int height)
     else glClearTexImage(id, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0);
 
     {GLuint64 handle = glGetTextureHandleARB(id);
-    glMakeTextureHandleResidentARB(handle);
-    dev->tex_handles[tex_index] = handle;
-    return tex_index;}
+    glMakeTextureHandleResidentARB(handle);}
+    return id;
 }
 
 NK_API void
-nk_glfw3_destroy_texture(int tex_index)
+nk_glfw3_destroy_texture(unsigned int id)
 {
-    struct nk_glfw_device *dev = &glfw.ogl;
-    GLuint id = nk_glfw3_get_tex_ogl_id(tex_index);
-    if (id == 0) return;
-
-    {GLuint64 handle = nk_glfw3_get_tex_ogl_handle(tex_index);
-    glMakeTextureHandleNonResidentARB(handle);
+    glMakeTextureHandleNonResidentARB(glGetTextureHandleARB(id));
     glDeleteTextures(1, &id);
-    dev->tex_ids[tex_index] = 0;
-    dev->tex_handles[tex_index] = 0;}
 }
 
 NK_INTERN void
@@ -313,7 +265,6 @@ nk_glfw3_device_upload_atlas(const void *image, int width, int height)
 NK_API void
 nk_glfw3_device_destroy(void)
 {
-    int i = 0;
     struct nk_glfw_device *dev = &glfw.ogl;
     glDetachShader(dev->prog, dev->vert_shdr);
     glDetachShader(dev->prog, dev->frag_shdr);
@@ -322,8 +273,6 @@ nk_glfw3_device_destroy(void)
     glDeleteProgram(dev->prog);
     nk_glfw3_destroy_texture(dev->font_tex_index);
 
-    for (i = 0; i < NK_GLFW_MAX_TEXTURES; i++)
-        nk_glfw3_destroy_texture(i);
     glUnmapNamedBuffer(dev->vbo);
     glUnmapNamedBuffer(dev->ebo);
     glDeleteBuffers(1, &dev->vbo);
@@ -407,7 +356,7 @@ nk_glfw3_render(enum nk_anti_aliasing AA)
                 config.vertex_layout = vertex_layout;
                 config.vertex_size = sizeof(struct nk_glfw_vertex);
                 config.vertex_alignment = NK_ALIGNOF(struct nk_glfw_vertex);
-                config.null = dev->null;
+                config.tex_null = dev->tex_null;
                 config.circle_segment_count = 22;
                 config.curve_segment_count = 22;
                 config.arc_segment_count = 22;
@@ -425,12 +374,10 @@ nk_glfw3_render(enum nk_anti_aliasing AA)
         /* iterate over and execute each draw command */
         nk_draw_foreach(cmd, &glfw.ctx, &dev->cmds)
         {
-            int tex_index;
             GLuint64 tex_handle;
             if (!cmd->elem_count) continue;
 
-            tex_index = cmd->texture.id;
-            tex_handle = nk_glfw3_get_tex_ogl_handle(tex_index);
+            tex_handle = glGetTextureHandleARB((unsigned int)cmd->texture.id);
 
             /* tex handle must be made resident in each context that uses it */
             if (!glIsTextureHandleResidentARB(tex_handle))
@@ -465,6 +412,56 @@ nk_glfw3_char_callback(GLFWwindow *win, unsigned int codepoint)
         glfw.text[glfw.text_len++] = codepoint;
 }
 
+NK_API void
+nk_glfw3_key_callback(GLFWwindow *win, int key, int scancode, int action, int mods)
+{
+    /*
+     * convert GLFW_REPEAT to down (technically GLFW_RELEASE, GLFW_PRESS, GLFW_REPEAT are
+     * already 0, 1, 2 but just to be clearer)
+     */
+    nk_char a = (action == GLFW_RELEASE) ? nk_false : nk_true;
+
+    NK_UNUSED(win);
+    NK_UNUSED(scancode);
+    NK_UNUSED(mods);
+
+    switch (key) {
+    case GLFW_KEY_DELETE:    glfw.key_events[NK_KEY_DEL] = a; break;
+    case GLFW_KEY_TAB:       glfw.key_events[NK_KEY_TAB] = a; break;
+    case GLFW_KEY_BACKSPACE: glfw.key_events[NK_KEY_BACKSPACE] = a; break;
+    case GLFW_KEY_UP:        glfw.key_events[NK_KEY_UP] = a; break;
+    case GLFW_KEY_DOWN:      glfw.key_events[NK_KEY_DOWN] = a; break;
+    case GLFW_KEY_LEFT:      glfw.key_events[NK_KEY_LEFT] = a; break;
+    case GLFW_KEY_RIGHT:     glfw.key_events[NK_KEY_RIGHT] = a; break;
+
+    case GLFW_KEY_PAGE_UP:   glfw.key_events[NK_KEY_SCROLL_UP] = a; break;
+    case GLFW_KEY_PAGE_DOWN: glfw.key_events[NK_KEY_SCROLL_DOWN] = a; break;
+
+    /* have to add all keys used for nuklear to get correct repeat behavior
+     * NOTE these are scancodes so your custom layout won't matter unfortunately
+     * Also while including everything will prevent unnecessary input calls,
+     * only the ones with visible effects really matter, ie paste, undo, redo
+     * selecting all, copying or cutting 40 times before you release the keys
+     * doesn't actually cause any visible problems */
+
+    case GLFW_KEY_C:         glfw.key_events[NK_KEY_COPY] = a; break;
+    case GLFW_KEY_V:         glfw.key_events[NK_KEY_PASTE] = a; break;
+    case GLFW_KEY_X:         glfw.key_events[NK_KEY_CUT] = a; break;
+    case GLFW_KEY_Z:         glfw.key_events[NK_KEY_TEXT_UNDO] = a; break;
+    case GLFW_KEY_R:         glfw.key_events[NK_KEY_TEXT_REDO] = a; break;
+    case GLFW_KEY_B:         glfw.key_events[NK_KEY_TEXT_LINE_START] = a; break;
+    case GLFW_KEY_E:         glfw.key_events[NK_KEY_TEXT_LINE_END] = a; break;
+    case GLFW_KEY_A:         glfw.key_events[NK_KEY_TEXT_SELECT_ALL] = a; break;
+
+    case GLFW_KEY_ENTER:
+    case GLFW_KEY_KP_ENTER:
+        glfw.key_events[NK_KEY_ENTER] = a;
+        break;
+    default:
+        ;
+    }
+}
+
 NK_API void
 nk_gflw3_scroll_callback(GLFWwindow *win, double xoff, double yoff)
 {
@@ -520,6 +517,7 @@ nk_glfw3_init(GLFWwindow *win, enum nk_glfw_init_state init_state,
     if (init_state == NK_GLFW3_INSTALL_CALLBACKS) {
         glfwSetScrollCallback(win, nk_gflw3_scroll_callback);
         glfwSetCharCallback(win, nk_glfw3_char_callback);
+        glfwSetKeyCallback(win, nk_glfw3_key_callback);
         glfwSetMouseButtonCallback(win, nk_glfw3_mouse_button_callback);
     }
     nk_init_default(&glfw.ctx, 0);
@@ -536,6 +534,8 @@ nk_glfw3_init(GLFWwindow *win, enum nk_glfw_init_state init_state,
     glfw.is_double_click_down = nk_false;
     glfw.double_click_pos = nk_vec2(0, 0);
 
+    glfw.delta_time_seconds_last = (float)glfwGetTime();
+
     return &glfw.ctx;
 }
 
@@ -553,7 +553,7 @@ nk_glfw3_font_stash_end(void)
     const void *image; int w, h;
     image = nk_font_atlas_bake(&glfw.atlas, &w, &h, NK_FONT_ATLAS_RGBA32);
     nk_glfw3_device_upload_atlas(image, w, h);
-    nk_font_atlas_end(&glfw.atlas, nk_handle_id((int)glfw.ogl.font_tex_index), &glfw.ogl.null);
+    nk_font_atlas_end(&glfw.atlas, nk_handle_id((int)glfw.ogl.font_tex_index), &glfw.ogl.tex_null);
     if (glfw.atlas.default_font)
         nk_style_set_font(&glfw.ctx, &glfw.atlas.default_font->handle);
 }
@@ -565,6 +565,12 @@ nk_glfw3_new_frame(void)
     double x, y;
     struct nk_context *ctx = &glfw.ctx;
     struct GLFWwindow *win = glfw.win;
+    nk_char* k_state = glfw.key_events;
+
+    /* update the timer */
+    float delta_time_now = (float)glfwGetTime();
+    glfw.ctx.delta_time_seconds = delta_time_now - glfw.delta_time_seconds_last;
+    glfw.delta_time_seconds_last = delta_time_now;
 
     glfwGetWindowSize(win, &glfw.width, &glfw.height);
     glfwGetFramebufferSize(win, &glfw.display_width, &glfw.display_height);
@@ -583,40 +589,42 @@ nk_glfw3_new_frame(void)
         glfwSetInputMode(glfw.win, GLFW_CURSOR, GLFW_CURSOR_NORMAL);
 #endif
 
-    nk_input_key(ctx, NK_KEY_DEL, glfwGetKey(win, GLFW_KEY_DELETE) == GLFW_PRESS);
-    nk_input_key(ctx, NK_KEY_ENTER, glfwGetKey(win, GLFW_KEY_ENTER) == GLFW_PRESS);
-    nk_input_key(ctx, NK_KEY_TAB, glfwGetKey(win, GLFW_KEY_TAB) == GLFW_PRESS);
-    nk_input_key(ctx, NK_KEY_BACKSPACE, glfwGetKey(win, GLFW_KEY_BACKSPACE) == GLFW_PRESS);
-    nk_input_key(ctx, NK_KEY_UP, glfwGetKey(win, GLFW_KEY_UP) == GLFW_PRESS);
-    nk_input_key(ctx, NK_KEY_DOWN, glfwGetKey(win, GLFW_KEY_DOWN) == GLFW_PRESS);
+    if (k_state[NK_KEY_DEL] >= 0) nk_input_key(ctx, NK_KEY_DEL, k_state[NK_KEY_DEL]);
+    if (k_state[NK_KEY_ENTER] >= 0) nk_input_key(ctx, NK_KEY_ENTER, k_state[NK_KEY_ENTER]);
+
+    if (k_state[NK_KEY_TAB] >= 0) nk_input_key(ctx, NK_KEY_TAB, k_state[NK_KEY_TAB]);
+    if (k_state[NK_KEY_BACKSPACE] >= 0) nk_input_key(ctx, NK_KEY_BACKSPACE, k_state[NK_KEY_BACKSPACE]);
+    if (k_state[NK_KEY_UP] >= 0) nk_input_key(ctx, NK_KEY_UP, k_state[NK_KEY_UP]);
+    if (k_state[NK_KEY_DOWN] >= 0) nk_input_key(ctx, NK_KEY_DOWN, k_state[NK_KEY_DOWN]);
+    if (k_state[NK_KEY_SCROLL_UP] >= 0) nk_input_key(ctx, NK_KEY_SCROLL_UP, k_state[NK_KEY_SCROLL_UP]);
+    if (k_state[NK_KEY_SCROLL_DOWN] >= 0) nk_input_key(ctx, NK_KEY_SCROLL_DOWN, k_state[NK_KEY_SCROLL_DOWN]);
+
     nk_input_key(ctx, NK_KEY_TEXT_START, glfwGetKey(win, GLFW_KEY_HOME) == GLFW_PRESS);
     nk_input_key(ctx, NK_KEY_TEXT_END, glfwGetKey(win, GLFW_KEY_END) == GLFW_PRESS);
     nk_input_key(ctx, NK_KEY_SCROLL_START, glfwGetKey(win, GLFW_KEY_HOME) == GLFW_PRESS);
     nk_input_key(ctx, NK_KEY_SCROLL_END, glfwGetKey(win, GLFW_KEY_END) == GLFW_PRESS);
-    nk_input_key(ctx, NK_KEY_SCROLL_DOWN, glfwGetKey(win, GLFW_KEY_PAGE_DOWN) == GLFW_PRESS);
-    nk_input_key(ctx, NK_KEY_SCROLL_UP, glfwGetKey(win, GLFW_KEY_PAGE_UP) == GLFW_PRESS);
     nk_input_key(ctx, NK_KEY_SHIFT, glfwGetKey(win, GLFW_KEY_LEFT_SHIFT) == GLFW_PRESS||
                                     glfwGetKey(win, GLFW_KEY_RIGHT_SHIFT) == GLFW_PRESS);
 
     if (glfwGetKey(win, GLFW_KEY_LEFT_CONTROL) == GLFW_PRESS ||
         glfwGetKey(win, GLFW_KEY_RIGHT_CONTROL) == GLFW_PRESS) {
-        nk_input_key(ctx, NK_KEY_COPY, glfwGetKey(win, GLFW_KEY_C) == GLFW_PRESS);
-        nk_input_key(ctx, NK_KEY_PASTE, glfwGetKey(win, GLFW_KEY_V) == GLFW_PRESS);
-        nk_input_key(ctx, NK_KEY_CUT, glfwGetKey(win, GLFW_KEY_X) == GLFW_PRESS);
-        nk_input_key(ctx, NK_KEY_TEXT_UNDO, glfwGetKey(win, GLFW_KEY_Z) == GLFW_PRESS);
-        nk_input_key(ctx, NK_KEY_TEXT_REDO, glfwGetKey(win, GLFW_KEY_R) == GLFW_PRESS);
-        nk_input_key(ctx, NK_KEY_TEXT_WORD_LEFT, glfwGetKey(win, GLFW_KEY_LEFT) == GLFW_PRESS);
-        nk_input_key(ctx, NK_KEY_TEXT_WORD_RIGHT, glfwGetKey(win, GLFW_KEY_RIGHT) == GLFW_PRESS);
-        nk_input_key(ctx, NK_KEY_TEXT_LINE_START, glfwGetKey(win, GLFW_KEY_B) == GLFW_PRESS);
-        nk_input_key(ctx, NK_KEY_TEXT_LINE_END, glfwGetKey(win, GLFW_KEY_E) == GLFW_PRESS);
-        nk_input_key(ctx, NK_KEY_TEXT_SELECT_ALL, glfwGetKey(win, GLFW_KEY_A) == GLFW_PRESS);
+        /* Note these are physical keys and won't respect any layouts/key mapping */
+        if (k_state[NK_KEY_COPY] >= 0) nk_input_key(ctx, NK_KEY_COPY, k_state[NK_KEY_COPY]);
+        if (k_state[NK_KEY_PASTE] >= 0) nk_input_key(ctx, NK_KEY_PASTE, k_state[NK_KEY_PASTE]);
+        if (k_state[NK_KEY_CUT] >= 0) nk_input_key(ctx, NK_KEY_CUT, k_state[NK_KEY_CUT]);
+        if (k_state[NK_KEY_TEXT_UNDO] >= 0) nk_input_key(ctx, NK_KEY_TEXT_UNDO, k_state[NK_KEY_TEXT_UNDO]);
+        if (k_state[NK_KEY_TEXT_REDO] >= 0) nk_input_key(ctx, NK_KEY_TEXT_REDO, k_state[NK_KEY_TEXT_REDO]);
+        if (k_state[NK_KEY_TEXT_LINE_START] >= 0) nk_input_key(ctx, NK_KEY_TEXT_LINE_START, k_state[NK_KEY_TEXT_LINE_START]);
+        if (k_state[NK_KEY_TEXT_LINE_END] >= 0) nk_input_key(ctx, NK_KEY_TEXT_LINE_END, k_state[NK_KEY_TEXT_LINE_END]);
+        if (k_state[NK_KEY_TEXT_SELECT_ALL] >= 0) nk_input_key(ctx, NK_KEY_TEXT_SELECT_ALL, k_state[NK_KEY_TEXT_SELECT_ALL]);
+        if (k_state[NK_KEY_LEFT] >= 0) nk_input_key(ctx, NK_KEY_TEXT_WORD_LEFT, k_state[NK_KEY_LEFT]);
+        if (k_state[NK_KEY_RIGHT] >= 0) nk_input_key(ctx, NK_KEY_TEXT_WORD_RIGHT, k_state[NK_KEY_RIGHT]);
     } else {
-        nk_input_key(ctx, NK_KEY_LEFT, glfwGetKey(win, GLFW_KEY_LEFT) == GLFW_PRESS);
-        nk_input_key(ctx, NK_KEY_RIGHT, glfwGetKey(win, GLFW_KEY_RIGHT) == GLFW_PRESS);
+        if (k_state[NK_KEY_LEFT] >= 0) nk_input_key(ctx, NK_KEY_LEFT, k_state[NK_KEY_LEFT]);
+        if (k_state[NK_KEY_RIGHT] >= 0) nk_input_key(ctx, NK_KEY_RIGHT, k_state[NK_KEY_RIGHT]);
         nk_input_key(ctx, NK_KEY_COPY, 0);
         nk_input_key(ctx, NK_KEY_PASTE, 0);
         nk_input_key(ctx, NK_KEY_CUT, 0);
-        nk_input_key(ctx, NK_KEY_SHIFT, 0);
     }
 
     glfwGetCursorPos(win, &x, &y);
@@ -634,6 +642,10 @@ nk_glfw3_new_frame(void)
     nk_input_button(ctx, NK_BUTTON_DOUBLE, (int)glfw.double_click_pos.x, (int)glfw.double_click_pos.y, glfw.is_double_click_down);
     nk_input_scroll(ctx, glfw.scroll);
     nk_input_end(&glfw.ctx);
+
+    /* clear after nk_input_end (-1 since we're doing up/down boolean) */
+    memset(glfw.key_events, -1, sizeof(glfw.key_events));
+
     glfw.text_len = 0;
     glfw.scroll = nk_vec2(0,0);
 }
