@@ -17,8 +17,10 @@
 
 #include "gpu.h"
 
+VK_CB_FUNC_DEF(vk_tex_deref);
+
 void vk_tex_barrier(pl_gpu gpu, struct vk_cmd *cmd, pl_tex tex,
-                    VkPipelineStageFlags stage, VkAccessFlags access,
+                    VkPipelineStageFlags2 stage, VkAccessFlags2 access,
                     VkImageLayout layout, uint32_t qf)
 {
     struct pl_vk *p = PL_PRIV(gpu);
@@ -39,16 +41,18 @@ void vk_tex_barrier(pl_gpu gpu, struct vk_cmd *cmd, pl_tex tex,
 
     struct vk_sync_scope last;
     bool is_trans = layout != tex_vk->layout, is_xfer = qf != tex_vk->qf;
-    last = vk_sem_barrier(vk, cmd, &tex_vk->sem, stage, access, is_trans || is_xfer);
+    last = vk_sem_barrier(cmd, &tex_vk->sem, stage, access, is_trans || is_xfer);
 
-    VkImageMemoryBarrier barr = {
-        .sType = VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER,
+    VkImageMemoryBarrier2 barr = {
+        .sType = VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER_2,
+        .srcStageMask = last.stage,
+        .srcAccessMask = last.access,
+        .dstStageMask = stage,
+        .dstAccessMask = access,
         .oldLayout = tex_vk->layout,
         .newLayout = layout,
         .srcQueueFamilyIndex = tex_vk->qf,
         .dstQueueFamilyIndex = qf,
-        .srcAccessMask = last.access,
-        .dstAccessMask = access,
         .image = tex_vk->img,
         .subresourceRange = {
             .aspectMask = tex_vk->aspect,
@@ -62,23 +66,29 @@ void vk_tex_barrier(pl_gpu gpu, struct vk_cmd *cmd, pl_tex tex,
         barr.oldLayout = VK_IMAGE_LAYOUT_UNDEFINED;
     }
 
+    if (tex_vk->ext_deps.num) {
+        // We need to guarantee that all external dependencies are satisfied
+        // before the barrier begins executing. The easiest way to ensure this
+        // is to add the stage mask at which we wait for the external dependency
+        // to the source stage mask of the image barrier.
+        barr.srcStageMask |= stage;
+    }
+
     if (last.access || is_trans || is_xfer) {
-        vk->CmdPipelineBarrier(cmd->buf, last.stage, stage, 0, 0, NULL,
-                               0, NULL, 1, &barr);
+        vk_cmd_barrier(cmd, &(VkDependencyInfo) {
+            .sType = VK_STRUCTURE_TYPE_DEPENDENCY_INFO,
+            .imageMemoryBarrierCount = 1,
+            .pImageMemoryBarriers = &barr,
+        });
     }
 
     tex_vk->qf = qf;
     tex_vk->layout = layout;
-    vk_cmd_callback(cmd, (vk_cb) vk_tex_deref, gpu, tex);
+    vk_cmd_callback(cmd, VK_CB_FUNC(vk_tex_deref), gpu, tex);
 
     for (int i = 0; i < tex_vk->ext_deps.num; i++)
         vk_cmd_dep(cmd, stage, tex_vk->ext_deps.elem[i]);
     tex_vk->ext_deps.num = 0;
-
-    if (tex_vk->ext_sync) {
-        vk_cmd_callback(cmd, (vk_cb) vk_sync_deref, gpu, tex_vk->ext_sync);
-        tex_vk->ext_sync = NULL;
-    }
 }
 
 static void vk_tex_destroy(pl_gpu gpu, struct pl_tex_t *tex)
@@ -90,8 +100,6 @@ static void vk_tex_destroy(pl_gpu gpu, struct pl_tex_t *tex)
     struct vk_ctx *vk = p->vk;
     struct pl_tex_vk *tex_vk = PL_PRIV(tex);
 
-    vk_sync_deref(gpu, tex_vk->ext_sync);
-    vk_sem_uninit(vk, &tex_vk->sem);
     vk->DestroyFramebuffer(vk->dev, tex_vk->framebuffer, PL_VK_ALLOC);
     vk->DestroyImageView(vk->dev, tex_vk->view, PL_VK_ALLOC);
     for (int i = 0; i < tex_vk->num_planes; i++)
@@ -128,8 +136,6 @@ static bool vk_init_image(pl_gpu gpu, pl_tex tex, pl_debug_tag debug_tag)
     pl_rc_init(&tex_vk->rc);
     if (tex_vk->num_planes)
         return true;
-    if (!vk_sem_init(vk, &tex_vk->sem, debug_tag))
-        return false;
     tex_vk->layout = VK_IMAGE_LAYOUT_UNDEFINED;
     tex_vk->transfer_queue = GRAPHICS;
     tex_vk->qf = VK_QUEUE_FAMILY_IGNORED; // will be set on first use, if needed
@@ -153,8 +159,19 @@ static bool vk_init_image(pl_gpu gpu, pl_tex tex, pl_debug_tag debug_tag)
             [VK_IMAGE_TYPE_3D] = VK_IMAGE_VIEW_TYPE_3D,
         };
 
+        static const VkFlags view_whitelist =
+            VK_IMAGE_USAGE_TRANSFER_SRC_BIT | VK_IMAGE_USAGE_TRANSFER_DST_BIT |
+            VK_IMAGE_USAGE_SAMPLED_BIT | VK_IMAGE_USAGE_STORAGE_BIT |
+            VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT | VK_IMAGE_USAGE_HOST_TRANSFER_BIT;
+
+        const VkImageViewUsageCreateInfo usage_info = {
+            .sType = VK_STRUCTURE_TYPE_IMAGE_VIEW_USAGE_CREATE_INFO,
+            .usage = tex_vk->usage_flags & view_whitelist,
+        };
+
         const VkImageViewCreateInfo vinfo = {
             .sType = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO,
+            .pNext = &usage_info,
             .image = tex_vk->img,
             .viewType = viewType[tex_vk->type],
             .format = tex_vk->img_fmt,
@@ -207,13 +224,13 @@ static bool vk_init_image(pl_gpu gpu, pl_tex tex, pl_debug_tag debug_tag)
             .layers = 1,
         };
 
-        if (finfo.width > vk->limits.maxFramebufferWidth ||
-            finfo.height > vk->limits.maxFramebufferHeight)
+        if (finfo.width > vk->props.limits.maxFramebufferWidth ||
+            finfo.height > vk->props.limits.maxFramebufferHeight)
         {
             PL_ERR(gpu, "Framebuffer of size %dx%d exceeds the maximum allowed "
                    "dimensions: %dx%d", finfo.width, finfo.height,
-                   vk->limits.maxFramebufferWidth,
-                   vk->limits.maxFramebufferHeight);
+                   vk->props.limits.maxFramebufferWidth,
+                   vk->props.limits.maxFramebufferHeight);
             goto error;
         }
 
@@ -267,20 +284,6 @@ pl_tex vk_tex_create(pl_gpu gpu, const struct pl_tex_params *params)
             goto error;
         }
 
-        // Statically check to see if we'd even be able to upload it at all
-        // and refuse right away if not. In theory, uploading can still fail
-        // based on the size of pl_tex_transfer_params.row_pitch, but for now
-        // this should be enough.
-        uint64_t texels = params->w * PL_DEF(params->h, 1) * PL_DEF(params->d, 1) *
-                          fmt->num_components;
-
-        if (texels > gpu->limits.max_buffer_texels) {
-            PL_ERR(gpu, "Failed creating texture with emulated texture format: "
-                   "texture dimensions exceed maximum texel buffer size! Try "
-                   "again with a different (non-emulated) format?");
-            goto error;
-        }
-
         // Our format emulation requires storage image support. In order to
         // make a bunch of checks happy, just mark it off as storable (and also
         // enable VK_IMAGE_USAGE_STORAGE_BIT, which we do below)
@@ -297,6 +300,8 @@ pl_tex vk_tex_create(pl_gpu gpu, const struct pl_tex_params *params)
     if ((params->blit_src || params->blit_dst) && tex_vk->num_planes)
         tex->params.storable = true;
 
+    const bool host_writable = params->host_writable || params->initial_data;
+
     VkImageUsageFlags usage = 0;
     VkImageCreateFlags flags = 0;
     if (tex->params.sampleable)
@@ -307,9 +312,17 @@ pl_tex vk_tex_create(pl_gpu gpu, const struct pl_tex_params *params)
         usage |= VK_IMAGE_USAGE_STORAGE_BIT;
     if (tex->params.host_readable || tex->params.blit_src)
         usage |= VK_IMAGE_USAGE_TRANSFER_SRC_BIT;
-    if (tex->params.host_writable || tex->params.blit_dst || params->initial_data)
+    if (host_writable || tex->params.blit_dst)
         usage |= VK_IMAGE_USAGE_TRANSFER_DST_BIT;
 
+    // Opportunistically try and use host image copies if possible
+    const VkDeviceSize size_estimate = fmt->internal_size * PL_ALIGN2(params->w, 8) *
+                                       PL_MAX(params->h, 1) * PL_MAX(params->d, 1);
+    if (((tex->params.host_readable && p->host_dl_layouts.num) ||
+         (host_writable && p->host_ul_layouts.num)) && fmtp->can_host_copy &&
+        (p->rebar_enabled || size_estimate < gpu->limits.max_mapped_vram / 1024))
+        usage |= VK_IMAGE_USAGE_HOST_TRANSFER_BIT_EXT;
+
     if (!usage) {
         // Vulkan requires images have at least *some* image usage set, but our
         // API is perfectly happy with a (useless) image. So just put
@@ -327,6 +340,7 @@ pl_tex vk_tex_create(pl_gpu gpu, const struct pl_tex_params *params)
     // will belong to, we're forced to share all of our images between all
     // command pools.
     uint32_t qfs[3] = {0};
+    pl_assert(vk->pools.num <= PL_ARRAY_SIZE(qfs));
     for (int i = 0; i < vk->pools.num; i++)
         qfs[i] = vk->pools.elem[i]->qf;
 
@@ -391,6 +405,7 @@ pl_tex vk_tex_create(pl_gpu gpu, const struct pl_tex_params *params)
         .export_handle = params->export_handle,
         .import_handle = params->import_handle,
         .shared_mem = params->shared_mem,
+        .debug_tag = params->debug_tag,
     };
 
     if (params->import_handle == PL_HANDLE_DMA_BUF) {
@@ -412,6 +427,7 @@ pl_tex vk_tex_create(pl_gpu gpu, const struct pl_tex_params *params)
 #endif
 
     if (params->export_handle == PL_HANDLE_DMA_BUF) {
+        pl_assert(drm_list.drmFormatModifierCount > 0);
         vk_link_struct(&iinfo, &drm_list);
         iinfo.tiling = VK_IMAGE_TILING_DRM_FORMAT_MODIFIER_EXT;
     }
@@ -419,7 +435,6 @@ pl_tex vk_tex_create(pl_gpu gpu, const struct pl_tex_params *params)
     // Double-check physical image format limits and fail if invalid
     VkPhysicalDeviceImageDrmFormatModifierInfoEXT drm_pinfo = {
         .sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_DRM_FORMAT_MODIFIER_INFO_EXT,
-        .drmFormatModifier = drm_explicit.drmFormatModifier,
         .sharingMode = iinfo.sharingMode,
         .queueFamilyIndexCount = iinfo.queueFamilyIndexCount,
         .pQueueFamilyIndices = iinfo.pQueueFamilyIndices,
@@ -430,8 +445,19 @@ pl_tex vk_tex_create(pl_gpu gpu, const struct pl_tex_params *params)
         .handleType = ext_info.handleTypes,
     };
 
-    if (handle_type == PL_HANDLE_DMA_BUF)
+    if (handle_type == PL_HANDLE_DMA_BUF) {
+        if (params->import_handle) {
+            // On import, we know exactly which format modifier to test
+            drm_pinfo.drmFormatModifier = drm_explicit.drmFormatModifier;
+        } else {
+            // On export, the choice of format modifier is ambiguous, because
+            // we offer the implementation a whole list to choose from. In
+            // principle, we must check *all* supported drm format modifiers,
+            // but in practice it should hopefully suffice to just check one
+            drm_pinfo.drmFormatModifier = drm_list.pDrmFormatModifiers[0];
+        }
         vk_link_struct(&ext_pinfo, &drm_pinfo);
+    }
 
     VkPhysicalDeviceImageFormatInfo2KHR pinfo = {
         .sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_IMAGE_FORMAT_INFO_2_KHR,
@@ -453,7 +479,7 @@ pl_tex vk_tex_create(pl_gpu gpu, const struct pl_tex_params *params)
     };
 
     VkResult res;
-    res = vk->GetPhysicalDeviceImageFormatProperties2KHR(vk->physd, &pinfo, &props);
+    res = vk->GetPhysicalDeviceImageFormatProperties2(vk->physd, &pinfo, &props);
     if (res == VK_ERROR_FORMAT_NOT_SUPPORTED) {
         PL_DEBUG(gpu, "Texture creation failed: not supported");
         goto error;
@@ -465,9 +491,9 @@ pl_tex vk_tex_create(pl_gpu gpu, const struct pl_tex_params *params)
     if (params->w > max.width || params->h > max.height || params->d > max.depth)
     {
         PL_ERR(gpu, "Requested image size %dx%dx%d exceeds the maximum allowed "
-               "dimensions %dx%dx%d for vulkan image format %x",
+               "dimensions %dx%dx%d for vulkan image format %s",
                params->w, params->h, params->d, max.width, max.height, max.depth,
-               (unsigned) iinfo.format);
+               vk_fmt_name(iinfo.format));
         goto error;
     }
 
@@ -676,8 +702,8 @@ void vk_tex_clear_ex(pl_gpu gpu, pl_tex tex, const union pl_clear_color color)
     if (!cmd)
         return;
 
-    vk_tex_barrier(gpu, cmd, tex, VK_PIPELINE_STAGE_TRANSFER_BIT,
-                   VK_ACCESS_TRANSFER_WRITE_BIT,
+    vk_tex_barrier(gpu, cmd, tex, VK_PIPELINE_STAGE_2_CLEAR_BIT,
+                   VK_ACCESS_2_TRANSFER_WRITE_BIT,
                    VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
                    VK_QUEUE_FAMILY_IGNORED);
 
@@ -709,10 +735,10 @@ void vk_tex_blit(pl_gpu gpu, const struct pl_tex_blit_params *params)
     bool planar_fallback = src_vk->aspect != VK_IMAGE_ASPECT_COLOR_BIT ||
                            dst_vk->aspect != VK_IMAGE_ASPECT_COLOR_BIT;
 
-    struct pl_rect3d src_rc = params->src_rc, dst_rc = params->dst_rc;
+    pl_rect3d src_rc = params->src_rc, dst_rc = params->dst_rc;
     bool requires_scaling = !pl_rect3d_eq(src_rc, dst_rc);
     if ((requires_scaling && blit_emulated) || planar_fallback) {
-        if (!pl_tex_blit_compute(gpu, p->dp, params))
+        if (!pl_tex_blit_compute(gpu, params))
             PL_ERR(gpu, "Failed emulating texture blit, incompatible textures?");
         return;
     }
@@ -721,19 +747,19 @@ void vk_tex_blit(pl_gpu gpu, const struct pl_tex_blit_params *params)
     if (!cmd)
         return;
 
-    vk_tex_barrier(gpu, cmd, params->src, VK_PIPELINE_STAGE_TRANSFER_BIT,
-                   VK_ACCESS_TRANSFER_READ_BIT,
-                   VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL,
-                   VK_QUEUE_FAMILY_IGNORED);
-
-    vk_tex_barrier(gpu, cmd, params->dst, VK_PIPELINE_STAGE_TRANSFER_BIT,
-                   VK_ACCESS_TRANSFER_WRITE_BIT,
-                   VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
-                   VK_QUEUE_FAMILY_IGNORED);
-
     // When the blit operation doesn't require scaling, we can use the more
     // efficient vkCmdCopyImage instead of vkCmdBlitImage
     if (!requires_scaling) {
+        vk_tex_barrier(gpu, cmd, params->src, VK_PIPELINE_STAGE_2_COPY_BIT,
+                       VK_ACCESS_2_TRANSFER_READ_BIT,
+                       VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL,
+                       VK_QUEUE_FAMILY_IGNORED);
+
+        vk_tex_barrier(gpu, cmd, params->dst, VK_PIPELINE_STAGE_2_COPY_BIT,
+                       VK_ACCESS_2_TRANSFER_WRITE_BIT,
+                       VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
+                       VK_QUEUE_FAMILY_IGNORED);
+
         pl_rect3d_normalize(&src_rc);
 
         VkImageCopy region = {
@@ -757,6 +783,16 @@ void vk_tex_blit(pl_gpu gpu, const struct pl_tex_blit_params *params)
         vk->CmdCopyImage(cmd->buf, src_vk->img, src_vk->layout,
                          dst_vk->img, dst_vk->layout, 1, &region);
     } else {
+        vk_tex_barrier(gpu, cmd, params->src, VK_PIPELINE_STAGE_2_BLIT_BIT,
+                       VK_ACCESS_2_TRANSFER_READ_BIT,
+                       VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL,
+                       VK_QUEUE_FAMILY_IGNORED);
+
+        vk_tex_barrier(gpu, cmd, params->dst, VK_PIPELINE_STAGE_2_BLIT_BIT,
+                       VK_ACCESS_2_TRANSFER_WRITE_BIT,
+                       VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
+                       VK_QUEUE_FAMILY_IGNORED);
+
         VkImageBlit region = {
             .srcSubresource = {
                 .aspectMask = src_vk->aspect,
@@ -836,6 +872,88 @@ static enum queue_type vk_img_copy_queue(pl_gpu gpu, pl_tex tex,
     }
 }
 
+static bool try_host_xfer(pl_gpu gpu, pl_tex tex, bool upload)
+{
+    struct pl_vk *p = PL_PRIV(gpu);
+    struct vk_ctx *vk = p->vk;
+    struct pl_tex_vk *tex_vk = PL_PRIV(tex);
+    const PL_ARRAY(VkImageLayout) *layouts = upload ? (void *) &p->host_ul_layouts
+                                                    : (void *) &p->host_dl_layouts;
+
+    if (!(tex_vk->usage_flags & VK_IMAGE_USAGE_HOST_TRANSFER_BIT_EXT) |
+        !layouts->num || tex->params.format->emulated)
+        return false;
+
+    // It's not worth stalling the GPU to wait for in-use textures; just emit
+    // these copy commands asynchronously
+    if (vk_tex_poll(gpu, tex, 0) || tex_vk->ext_deps.num)
+        return false;
+
+    for (int i = 0; i < layouts->num; i++) {
+        if (tex_vk->layout == layouts->elem[i])
+            return true;
+    }
+
+    // Need to perform a layout transition
+    VK(vk->TransitionImageLayoutEXT(vk->dev, 1, &(VkHostImageLayoutTransitionInfoEXT) {
+        .sType = VK_STRUCTURE_TYPE_HOST_IMAGE_LAYOUT_TRANSITION_INFO_EXT,
+        .image = tex_vk->img,
+        .oldLayout = tex_vk->layout,
+        .newLayout = VK_IMAGE_LAYOUT_GENERAL,
+        .subresourceRange = {
+            .aspectMask = tex_vk->aspect,
+            .levelCount = 1,
+            .layerCount = 1,
+        },
+    }));
+    tex_vk->layout = VK_IMAGE_LAYOUT_GENERAL;
+    return true;
+
+error:
+    return false;
+}
+
+static bool vk_tex_upload_host(pl_gpu gpu, const struct pl_tex_transfer_params *params)
+{
+    struct pl_vk *p = PL_PRIV(gpu);
+    struct vk_ctx *vk = p->vk;
+    pl_rect3d rc = params->rc;
+    pl_tex tex = params->tex;
+    pl_fmt fmt = tex->params.format;
+    struct pl_tex_vk *tex_vk = PL_PRIV(tex);
+
+    if (!try_host_xfer(gpu, tex, true))
+        return pl_tex_upload_pbo(gpu, params);
+
+    VK(vk->CopyMemoryToImageEXT(vk->dev, &(VkCopyMemoryToImageInfo) {
+        .sType = VK_STRUCTURE_TYPE_COPY_MEMORY_TO_IMAGE_INFO_EXT,
+        .dstImage = tex_vk->img,
+        .dstImageLayout = tex_vk->layout,
+        .regionCount = 1,
+        .pRegions = &(VkMemoryToImageCopy) {
+            .sType = VK_STRUCTURE_TYPE_MEMORY_TO_IMAGE_COPY,
+            .pHostPointer = params->ptr,
+            .memoryRowLength = params->row_pitch / fmt->texel_size,
+            .memoryImageHeight = params->depth_pitch / params->row_pitch,
+            .imageOffset = { rc.x0, rc.y0, rc.z0 },
+            .imageExtent = { rc.x1, rc.y1, rc.z1 },
+            .imageSubresource = {
+                .aspectMask = tex_vk->aspect,
+                .layerCount = 1,
+            },
+        },
+    }));
+
+    tex_vk->may_invalidate = false;
+
+    if (params->callback)
+        params->callback(params->priv);
+    return true;
+
+error:
+    return false;
+}
+
 static void tex_xfer_cb(void *ctx, void *arg)
 {
     void (*fun)(void *priv) = ctx;
@@ -849,81 +967,83 @@ bool vk_tex_upload(pl_gpu gpu, const struct pl_tex_transfer_params *params)
     pl_tex tex = params->tex;
     pl_fmt fmt = tex->params.format;
     struct pl_tex_vk *tex_vk = PL_PRIV(tex);
+    struct pl_tex_transfer_params *slices = NULL;
+    int num_slices = 0;
 
     if (!params->buf)
-        return pl_tex_upload_pbo(gpu, params);
+        return vk_tex_upload_host(gpu, params);
 
     pl_buf buf = params->buf;
     struct pl_buf_vk *buf_vk = PL_PRIV(buf);
-    struct pl_rect3d rc = params->rc;
-    size_t size = pl_tex_transfer_size(params);
-
-    size_t buf_offset = buf_vk->mem.offset + params->buf_offset;
+    pl_rect3d rc = params->rc;
+    const size_t size = pl_tex_transfer_size(params);
+    const size_t buf_offset = buf_vk->mem.offset + params->buf_offset;
     bool unaligned = buf_offset % fmt->texel_size;
     if (unaligned)
         PL_TRACE(gpu, "vk_tex_upload: unaligned transfer (slow path)");
 
     if (fmt->emulated || unaligned) {
 
-        // Copy the source data buffer into an intermediate buffer
-        struct pl_buf_params tbuf_params = {
-            .debug_tag = PL_DEBUG_TAG,
-            .memory_type = PL_BUF_MEM_DEVICE,
-            .format = tex_vk->texel_fmt,
-            .size = size,
-        };
+        // Create all slice buffers first, to early-fail if OOM, and to avoid
+        // blocking unnecessarily on waiting for these buffers to get read from
+        num_slices = pl_tex_transfer_slices(gpu, tex_vk->texel_fmt, params, &slices);
+        for (int i = 0; i < num_slices; i++) {
+            slices[i].buf = pl_buf_create(gpu, pl_buf_params(
+                .memory_type = PL_BUF_MEM_DEVICE,
+                .format      = tex_vk->texel_fmt,
+                .size        = pl_tex_transfer_size(&slices[i]),
+                .storable    = fmt->emulated,
+            ));
 
-        if (fmt->emulated) {
-            if (size <= gpu->limits.max_ubo_size) {
-                tbuf_params.uniform = true;
-            } else if (size <= gpu->limits.max_ssbo_size) {
-                tbuf_params.storable = true;
-            } else {
-                // TODO: Implement strided upload path if really necessary
-                PL_ERR(gpu, "Texel buffer size requirements exceed GPU "
-                       "capabilities, failed uploading!");
+            if (!slices[i].buf) {
+                PL_ERR(gpu, "Failed creating buffer for tex upload fallback!");
+                num_slices = i; // only clean up buffers up to here
                 goto error;
             }
         }
 
-        pl_buf tbuf = pl_buf_create(gpu, &tbuf_params);
-        if (!tbuf) {
-            PL_ERR(gpu, "Failed creating buffer for tex upload fallback!");
-            goto error;
-        }
-
-        struct vk_cmd *cmd = CMD_BEGIN_TIMED(tex_vk->transfer_queue, params->timer);
+        // All temporary buffers successfully created, begin copying source data
+        struct vk_cmd *cmd = CMD_BEGIN_TIMED(tex_vk->transfer_queue,
+                                             params->timer);
         if (!cmd)
             goto error;
 
-        struct pl_buf_vk *tbuf_vk = PL_PRIV(tbuf);
-        VkBufferCopy region = {
-            .srcOffset = buf_offset,
-            .dstOffset = tbuf_vk->mem.offset,
-            .size = size,
-        };
-
-        vk_buf_barrier(gpu, cmd, buf, VK_PIPELINE_STAGE_TRANSFER_BIT,
-                       VK_ACCESS_TRANSFER_READ_BIT, params->buf_offset, size,
+        vk_buf_barrier(gpu, cmd, buf, VK_PIPELINE_STAGE_2_COPY_BIT,
+                       VK_ACCESS_2_TRANSFER_READ_BIT, params->buf_offset, size,
                        false);
-        vk_buf_barrier(gpu, cmd, tbuf, VK_PIPELINE_STAGE_TRANSFER_BIT,
-                       VK_ACCESS_TRANSFER_WRITE_BIT, 0, size, false);
-        vk->CmdCopyBuffer(cmd->buf, buf_vk->mem.buf, tbuf_vk->mem.buf,
-                          1, &region);
+
+        for (int i = 0; i < num_slices; i++) {
+            pl_buf slice = slices[i].buf;
+            struct pl_buf_vk *slice_vk = PL_PRIV(slice);
+            vk_buf_barrier(gpu, cmd, slice, VK_PIPELINE_STAGE_2_COPY_BIT,
+                           VK_ACCESS_2_TRANSFER_WRITE_BIT, 0, slice->params.size,
+                           false);
+
+            vk->CmdCopyBuffer(cmd->buf, buf_vk->mem.buf, slice_vk->mem.buf, 1, &(VkBufferCopy) {
+                .srcOffset = buf_vk->mem.offset + slices[i].buf_offset,
+                .dstOffset = slice_vk->mem.offset,
+                .size      = slice->params.size,
+            });
+        }
 
         if (params->callback)
             vk_cmd_callback(cmd, tex_xfer_cb, params->callback, params->priv);
 
-        CMD_FINISH(&cmd);
+        bool ok = CMD_FINISH(&cmd);
 
-        struct pl_tex_transfer_params fixed = *params;
-        fixed.buf = tbuf;
-        fixed.buf_offset = 0;
+        // Finally, dispatch the (texel) upload asynchronously. We can fire
+        // the callback already at the completion of previous command because
+        // these temporary buffers already hold persistent copies of the data
+        for (int i = 0; i < num_slices; i++) {
+            if (ok) {
+                slices[i].buf_offset = 0;
+                ok = fmt->emulated ? pl_tex_upload_texel(gpu, &slices[i])
+                                   : pl_tex_upload(gpu, &slices[i]);
+            }
+            pl_buf_destroy(gpu, &slices[i].buf);
+        }
 
-        bool ok = fmt->emulated ? pl_tex_upload_texel(gpu, p->dp, &fixed)
-                                : pl_tex_upload(gpu, &fixed);
-
-        pl_buf_destroy(gpu, &tbuf);
+        pl_free(slices);
         return ok;
 
     } else {
@@ -946,11 +1066,11 @@ bool vk_tex_upload(pl_gpu gpu, const struct pl_tex_transfer_params *params)
         if (!cmd)
             goto error;
 
-        vk_buf_barrier(gpu, cmd, buf, VK_PIPELINE_STAGE_TRANSFER_BIT,
-                       VK_ACCESS_TRANSFER_READ_BIT, params->buf_offset, size,
+        vk_buf_barrier(gpu, cmd, buf, VK_PIPELINE_STAGE_2_COPY_BIT,
+                       VK_ACCESS_2_TRANSFER_READ_BIT, params->buf_offset, size,
                        false);
-        vk_tex_barrier(gpu, cmd, tex, VK_PIPELINE_STAGE_TRANSFER_BIT,
-                       VK_ACCESS_TRANSFER_WRITE_BIT,
+        vk_tex_barrier(gpu, cmd, tex, VK_PIPELINE_STAGE_2_COPY_BIT,
+                       VK_ACCESS_2_TRANSFER_WRITE_BIT,
                        VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL,
                        VK_QUEUE_FAMILY_IGNORED);
         vk->CmdCopyBufferToImage(cmd->buf, buf_vk->mem.buf, tex_vk->img,
@@ -959,12 +1079,15 @@ bool vk_tex_upload(pl_gpu gpu, const struct pl_tex_transfer_params *params)
         if (params->callback)
             vk_cmd_callback(cmd, tex_xfer_cb, params->callback, params->priv);
 
-        CMD_FINISH(&cmd);
+        return CMD_FINISH(&cmd);
     }
 
-    return true;
+    pl_unreachable();
 
 error:
+    for (int i = 0; i < num_slices; i++)
+        pl_buf_destroy(gpu, &slices[i].buf);
+    pl_free(slices);
     return false;
 }
 
@@ -975,78 +1098,89 @@ bool vk_tex_download(pl_gpu gpu, const struct pl_tex_transfer_params *params)
     pl_tex tex = params->tex;
     pl_fmt fmt = tex->params.format;
     struct pl_tex_vk *tex_vk = PL_PRIV(tex);
+    struct pl_tex_transfer_params *slices = NULL;
+    int num_slices = 0;
 
     if (!params->buf)
         return pl_tex_download_pbo(gpu, params);
 
     pl_buf buf = params->buf;
     struct pl_buf_vk *buf_vk = PL_PRIV(buf);
-    struct pl_rect3d rc = params->rc;
-    size_t size = pl_tex_transfer_size(params);
-
-    size_t buf_offset = buf_vk->mem.offset + params->buf_offset;
+    pl_rect3d rc = params->rc;
+    const size_t size = pl_tex_transfer_size(params);
+    const size_t buf_offset = buf_vk->mem.offset + params->buf_offset;
     bool unaligned = buf_offset % fmt->texel_size;
     if (unaligned)
         PL_TRACE(gpu, "vk_tex_download: unaligned transfer (slow path)");
 
     if (fmt->emulated || unaligned) {
 
-        // Download into an intermediate buffer first
-        pl_buf tbuf = pl_buf_create(gpu, pl_buf_params(
-            .storable = fmt->emulated,
-            .size = size,
-            .memory_type = PL_BUF_MEM_DEVICE,
-            .format = tex_vk->texel_fmt,
-        ));
+        num_slices = pl_tex_transfer_slices(gpu, tex_vk->texel_fmt, params, &slices);
+        for (int i = 0; i < num_slices; i++) {
+            slices[i].buf = pl_buf_create(gpu, pl_buf_params(
+                .memory_type = PL_BUF_MEM_DEVICE,
+                .format      = tex_vk->texel_fmt,
+                .size        = pl_tex_transfer_size(&slices[i]),
+                .storable    = fmt->emulated,
+            ));
 
-        if (!tbuf) {
-            PL_ERR(gpu, "Failed creating buffer for tex download fallback!");
-            goto error;
+            if (!slices[i].buf) {
+                PL_ERR(gpu, "Failed creating buffer for tex download fallback!");
+                num_slices = i;
+                goto error;
+            }
         }
 
-        struct pl_tex_transfer_params fixed = *params;
-        fixed.buf = tbuf;
-        fixed.buf_offset = 0;
-
-        bool ok = fmt->emulated ? pl_tex_download_texel(gpu, p->dp, &fixed)
-                                : pl_tex_download(gpu, &fixed);
-        if (!ok) {
-            pl_buf_destroy(gpu, &tbuf);
-            goto error;
+        for (int i = 0; i < num_slices; i++) {
+            // Restore buffer offset after downloading into temporary buffer,
+            // because we still need to copy the data from the temporary buffer
+            // into this offset in the original buffer
+            const size_t tmp_offset = slices[i].buf_offset;
+            slices[i].buf_offset = 0;
+            bool ok = fmt->emulated ? pl_tex_download_texel(gpu, &slices[i])
+                                    : pl_tex_download(gpu, &slices[i]);
+            slices[i].buf_offset = tmp_offset;
+            if (!ok)
+                goto error;
         }
 
+        // Finally, download into the user buffer
         struct vk_cmd *cmd = CMD_BEGIN_TIMED(tex_vk->transfer_queue, params->timer);
-        if (!cmd) {
-            pl_buf_destroy(gpu, &tbuf);
+        if (!cmd)
             goto error;
+
+        vk_buf_barrier(gpu, cmd, buf, VK_PIPELINE_STAGE_2_COPY_BIT,
+                       VK_ACCESS_2_TRANSFER_WRITE_BIT, params->buf_offset, size,
+                       false);
+
+        for (int i = 0; i < num_slices; i++) {
+            pl_buf slice = slices[i].buf;
+            struct pl_buf_vk *slice_vk = PL_PRIV(slice);
+            vk_buf_barrier(gpu, cmd, slice, VK_PIPELINE_STAGE_2_COPY_BIT,
+                           VK_ACCESS_2_TRANSFER_READ_BIT, 0, slice->params.size,
+                           false);
+
+            vk->CmdCopyBuffer(cmd->buf, slice_vk->mem.buf, buf_vk->mem.buf, 1, &(VkBufferCopy) {
+                .srcOffset = slice_vk->mem.offset,
+                .dstOffset = buf_vk->mem.offset + slices[i].buf_offset,
+                .size      = slice->params.size,
+            });
+
+            pl_buf_destroy(gpu, &slices[i].buf);
         }
 
-        struct pl_buf_vk *tbuf_vk = PL_PRIV(tbuf);
-        VkBufferCopy region = {
-            .srcOffset = tbuf_vk->mem.offset,
-            .dstOffset = buf_offset,
-            .size = size,
-        };
-
-        vk_buf_barrier(gpu, cmd, tbuf, VK_PIPELINE_STAGE_TRANSFER_BIT,
-                       VK_ACCESS_TRANSFER_READ_BIT, 0, size, false);
-        vk_buf_barrier(gpu, cmd, buf, VK_PIPELINE_STAGE_TRANSFER_BIT,
-                       VK_ACCESS_TRANSFER_WRITE_BIT, params->buf_offset, size,
-                       false);
-        vk->CmdCopyBuffer(cmd->buf, tbuf_vk->mem.buf, buf_vk->mem.buf,
-                          1, &region);
         vk_buf_flush(gpu, cmd, buf, params->buf_offset, size);
 
         if (params->callback)
             vk_cmd_callback(cmd, tex_xfer_cb, params->callback, params->priv);
 
-
-        CMD_FINISH(&cmd);
-        pl_buf_destroy(gpu, &tbuf);
+        pl_free(slices);
+        return CMD_FINISH(&cmd);
 
     } else {
 
-        pl_assert(fmt->texel_align == fmt->texel_size);
+        pl_assert(params->row_pitch % fmt->texel_size == 0);
+        pl_assert(params->depth_pitch % params->row_pitch == 0);
         const VkBufferImageCopy region = {
             .bufferOffset = buf_offset,
             .bufferRowLength = params->row_pitch / fmt->texel_size,
@@ -1065,11 +1199,11 @@ bool vk_tex_download(pl_gpu gpu, const struct pl_tex_transfer_params *params)
         if (!cmd)
             goto error;
 
-        vk_buf_barrier(gpu, cmd, buf, VK_PIPELINE_STAGE_TRANSFER_BIT,
-                       VK_ACCESS_TRANSFER_WRITE_BIT, params->buf_offset, size,
+        vk_buf_barrier(gpu, cmd, buf, VK_PIPELINE_STAGE_2_COPY_BIT,
+                       VK_ACCESS_2_TRANSFER_WRITE_BIT, params->buf_offset, size,
                        false);
-        vk_tex_barrier(gpu, cmd, tex, VK_PIPELINE_STAGE_TRANSFER_BIT,
-                       VK_ACCESS_TRANSFER_READ_BIT,
+        vk_tex_barrier(gpu, cmd, tex, VK_PIPELINE_STAGE_2_COPY_BIT,
+                       VK_ACCESS_2_TRANSFER_READ_BIT,
                        VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL,
                        VK_QUEUE_FAMILY_IGNORED);
         vk->CmdCopyImageToBuffer(cmd->buf, tex_vk->img, tex_vk->layout,
@@ -1079,13 +1213,15 @@ bool vk_tex_download(pl_gpu gpu, const struct pl_tex_transfer_params *params)
         if (params->callback)
             vk_cmd_callback(cmd, tex_xfer_cb, params->callback, params->priv);
 
-
-        CMD_FINISH(&cmd);
+        return CMD_FINISH(&cmd);
     }
 
-    return true;
+    pl_unreachable();
 
 error:
+    for (int i = 0; i < num_slices; i++)
+        pl_buf_destroy(gpu, &slices[i].buf);
+    pl_free(slices);
     return false;
 }
 
@@ -1117,43 +1253,6 @@ skip_blocking:
     return false;
 }
 
-bool vk_tex_export(pl_gpu gpu, pl_tex tex, pl_sync sync)
-{
-    struct pl_tex_vk *tex_vk = PL_PRIV(tex);
-    struct pl_sync_vk *sync_vk = PL_PRIV(sync);
-
-    if (tex_vk->num_planes) {
-        PL_ERR(gpu, "`pl_tex_export` cannot be called on planar textures."
-               "Please see `pl_vulkan_hold_ex` for a replacement.");
-        return false;
-    }
-
-    struct vk_cmd *cmd = CMD_BEGIN(ANY);
-    if (!cmd)
-        goto error;
-
-    vk_tex_barrier(gpu, cmd, tex, VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT,
-                   0, VK_IMAGE_LAYOUT_GENERAL, VK_QUEUE_FAMILY_EXTERNAL);
-
-    // Make the next barrier appear as though coming from a different queue
-    tex_vk->sem.write.queue = tex_vk->sem.read.queue = NULL;
-
-    vk_cmd_sig(cmd, (pl_vulkan_sem){ sync_vk->wait });
-    if (!CMD_SUBMIT(&cmd))
-        goto error;
-
-    // Remember the other dependency and hold on to the sync object
-    PL_ARRAY_APPEND(tex, tex_vk->ext_deps, (pl_vulkan_sem){ sync_vk->signal });
-    pl_rc_ref(&sync_vk->rc);
-    tex_vk->ext_sync = sync;
-    tex_vk->qf = VK_QUEUE_FAMILY_EXTERNAL;
-    return true;
-
-error:
-    PL_ERR(gpu, "Failed exporting shared texture!");
-    return false;
-}
-
 pl_tex pl_vulkan_wrap(pl_gpu gpu, const struct pl_vulkan_wrap_params *params)
 {
     pl_fmt fmt = NULL;
@@ -1230,7 +1329,13 @@ pl_tex pl_vulkan_wrap(pl_gpu gpu, const struct pl_vulkan_wrap_params *params)
     tex_vk->img_fmt = params->format;
     tex_vk->num_planes = fmt->num_planes;
     tex_vk->usage_flags = usage;
-    tex_vk->aspect = PL_DEF(params->aspect, VK_IMAGE_ASPECT_COLOR_BIT);
+    tex_vk->aspect = params->aspect;
+
+    if (!tex_vk->aspect) {
+        for (int i = 0; i < tex_vk->num_planes; i++)
+            tex_vk->aspect |= VK_IMAGE_ASPECT_PLANE_0_BIT << i;
+        tex_vk->aspect = PL_DEF(tex_vk->aspect, VK_IMAGE_ASPECT_COLOR_BIT);
+    }
 
     // Blitting to planar images requires fallback via compute shaders
     if (tex_vk->aspect != VK_IMAGE_ASPECT_COLOR_BIT) {
@@ -1245,10 +1350,10 @@ pl_tex pl_vulkan_wrap(pl_gpu gpu, const struct pl_vulkan_wrap_params *params)
     for (int i = 0; i < tex_vk->num_planes; i++) {
         struct pl_tex_t *plane;
         VkImageAspectFlags aspect = VK_IMAGE_ASPECT_PLANE_0_BIT << i;
-        if (!(aspect & params->aspect)) {
+        if (!(aspect & tex_vk->aspect)) {
             PL_INFO(gpu, "Not wrapping plane %d due to aspect bit 0x%x not "
                     "being contained in supplied params->aspect 0x%x!",
-                    i, (unsigned) aspect, (unsigned) params->aspect);
+                    i, (unsigned) aspect, (unsigned) tex_vk->aspect);
             continue;
         }
 
@@ -1328,19 +1433,17 @@ bool pl_vulkan_hold_ex(pl_gpu gpu, const struct pl_vulkan_hold_params *params)
     bool may_invalidate = true;
     if (!tex_vk->num_planes) {
         may_invalidate &= tex_vk->may_invalidate;
-        vk_tex_barrier(gpu, cmd, params->tex,
-                       VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT,
+        vk_tex_barrier(gpu, cmd, params->tex, VK_PIPELINE_STAGE_2_NONE,
                        0, layout, params->qf);
     }
 
     for (int i = 0; i < tex_vk->num_planes; i++) {
         may_invalidate &= tex_vk->planes[i]->may_invalidate;
         vk_tex_barrier(gpu, cmd, params->tex->planes[i],
-                       VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT,
-                       0, layout, params->qf);
+                       VK_PIPELINE_STAGE_2_NONE, 0, layout, params->qf);
     }
 
-    vk_cmd_sig(cmd, params->semaphore);
+    vk_cmd_sig(cmd, VK_PIPELINE_STAGE_2_ALL_COMMANDS_BIT, params->semaphore);
     bool ok = CMD_SUBMIT(&cmd);
 
     if (!tex_vk->num_planes) {
