@@ -23,21 +23,22 @@
 #include "swapchain.h"
 #include "pl_thread.h"
 
-struct sem_pair {
-    VkSemaphore in;
-    VkSemaphore out;
-};
-
 struct priv {
+    struct pl_sw_fns impl;
+
     pl_mutex lock;
     struct vk_ctx *vk;
     VkSurfaceKHR surf;
     PL_ARRAY(VkSurfaceFormatKHR) formats;
 
+    // true if host implementation has color management support enabled
+    bool have_color_management_support;
+
     // current swapchain and metadata:
     struct pl_vulkan_swapchain_params params;
     VkSwapchainCreateInfoKHR protoInfo; // partially filled-in prototype
     VkSwapchainKHR swapchain;
+    uint32_t queue_families[3];
     int cur_width, cur_height;
     int swapchain_depth;
     pl_rc_t frames_in_flight;       // number of frames currently queued
@@ -49,44 +50,48 @@ struct priv {
 
     // state of the images:
     PL_ARRAY(pl_tex) images;        // pl_tex wrappers for the VkImages
-    PL_ARRAY(struct sem_pair) sems; // pool of semaphores used to synchronize images
-    int idx_sems;                   // index of next free semaphore pair
+    PL_ARRAY(VkSemaphore) sems_in;  // pool of semaphores used to acquire images
+    PL_ARRAY(VkSemaphore) sems_out; // pool of semaphores used to present images
+    int idx_sems_in;                // index of next free semaphore to acquire
     int last_imgidx;                // the image index last acquired (for submit)
 };
 
-static struct pl_sw_fns vulkan_swapchain;
+static const struct pl_sw_fns vulkan_swapchain;
 
 static bool map_color_space(VkColorSpaceKHR space, struct pl_color_space *out)
 {
     switch (space) {
-    // Note: This is technically against the spec, but more often than not
-    // it's the correct result since `SRGB_NONLINEAR` is just a catch-all
-    // for any sort of typical SDR curve, which is better approximated by
-    // `pl_color_space_monitor`.
     case VK_COLOR_SPACE_SRGB_NONLINEAR_KHR:
-        *out = pl_color_space_monitor;
+        *out = (struct pl_color_space) {
+            .primaries = PL_COLOR_PRIM_BT_709,
+            .transfer  = PL_COLOR_TRC_SRGB,
+        };
         return true;
-
-#ifdef VK_EXT_swapchain_colorspace
     case VK_COLOR_SPACE_BT709_NONLINEAR_EXT:
-        *out = pl_color_space_monitor;
+        *out = (struct pl_color_space) {
+            .primaries = PL_COLOR_PRIM_BT_709,
+            .transfer  = PL_COLOR_TRC_BT_1886,
+        };
         return true;
     case VK_COLOR_SPACE_DISPLAY_P3_NONLINEAR_EXT:
         *out = (struct pl_color_space) {
             .primaries = PL_COLOR_PRIM_DISPLAY_P3,
-            .transfer  = PL_COLOR_TRC_BT_1886,
+            // Actually there are some controversy about Display P3's TRC curve,
+            // just like sRGB
+            .transfer  = PL_COLOR_TRC_SRGB,
         };
         return true;
-    case VK_COLOR_SPACE_DCI_P3_LINEAR_EXT:
+    case VK_COLOR_SPACE_DISPLAY_P3_LINEAR_EXT:
         *out = (struct pl_color_space) {
-            .primaries = PL_COLOR_PRIM_DCI_P3,
+            .primaries = PL_COLOR_PRIM_DISPLAY_P3,
             .transfer  = PL_COLOR_TRC_LINEAR,
         };
         return true;
     case VK_COLOR_SPACE_DCI_P3_NONLINEAR_EXT:
+        // This color space is using XYZ color system than RGB
         *out = (struct pl_color_space) {
             .primaries = PL_COLOR_PRIM_DCI_P3,
-            .transfer  = PL_COLOR_TRC_BT_1886,
+            .transfer  = PL_COLOR_TRC_ST428,
         };
         return true;
     case VK_COLOR_SPACE_EXTENDED_SRGB_LINEAR_EXT:
@@ -95,7 +100,7 @@ static bool map_color_space(VkColorSpaceKHR space, struct pl_color_space *out)
         return false;
     case VK_COLOR_SPACE_BT709_LINEAR_EXT:
         *out = (struct pl_color_space) {
-            .primaries = PL_COLOR_PRIM_DCI_P3,
+            .primaries = PL_COLOR_PRIM_BT_709,
             .transfer  = PL_COLOR_TRC_LINEAR,
         };
         return true;
@@ -133,9 +138,11 @@ static bool map_color_space(VkColorSpaceKHR space, struct pl_color_space *out)
         };
         return true;
     case VK_COLOR_SPACE_PASS_THROUGH_EXT:
-        *out = pl_color_space_unknown;
-        return true;
-#endif
+        // On color-managed Wayland compositors, it behaves similar to
+        // VK_COLOR_SPACE_SRGB_NONLINEAR_KHR as they would treat un-tagged surface
+        // as sRGB, but on other OSes it's behavior is not clearly defined, so
+        // don't use it.
+        return false;
 
 #ifdef VK_AMD_display_native_hdr
     case VK_COLOR_SPACE_DISPLAY_NATIVE_AMD:
@@ -163,6 +170,9 @@ static bool pick_surf_format(pl_swapchain sw, const struct pl_color_space *hint)
         if (!map_color_space(p->formats.elem[i].colorSpace, &space))
             continue;
 
+        bool disable10 = !pl_color_transfer_is_hdr(space.transfer) &&
+                         p->params.disable_10bit_sdr;
+
         switch (p->formats.elem[i].format) {
         // Only accept floating point formats for linear curves
         case VK_FORMAT_R16G16B16_SFLOAT:
@@ -188,14 +198,16 @@ static bool pick_surf_format(pl_swapchain sw, const struct pl_color_space *hint)
         // Only accept 10 bit formats for non-linear curves
         case VK_FORMAT_A2R10G10B10_UNORM_PACK32:
         case VK_FORMAT_A2B10G10R10_UNORM_PACK32:
-            if (space.transfer != PL_COLOR_TRC_LINEAR)
+            if (space.transfer != PL_COLOR_TRC_LINEAR && !disable10)
                 break; // accept
             continue;
 
         // Accept 16-bit formats for everything
         case VK_FORMAT_R16G16B16_UNORM:
         case VK_FORMAT_R16G16B16A16_UNORM:
-             break; // accept
+            if (!disable10)
+                break; // accept
+            continue;
 
         default: continue;
         }
@@ -219,8 +231,10 @@ static bool pick_surf_format(pl_swapchain sw, const struct pl_color_space *hint)
                 score += plfmt->component_depth[c];
             if (pl_color_primaries_is_wide_gamut(space.primaries) == wide_gamut)
                 score += 1000;
-            if (space.primaries == hint->primaries)
+            if (space.transfer == PL_COLOR_TRC_LINEAR)
                 score += 2000;
+            if (space.primaries == hint->primaries)
+                score += 4000;
             if (pl_color_transfer_is_hdr(space.transfer) == prefer_hdr)
                 score += 10000;
             if (space.transfer == hint->transfer)
@@ -230,9 +244,9 @@ static bool pick_surf_format(pl_swapchain sw, const struct pl_color_space *hint)
             case PL_FMT_UNKNOWN: break;
             case PL_FMT_UINT: break;
             case PL_FMT_SINT: break;
-            case PL_FMT_UNORM: score += 500; break;
-            case PL_FMT_SNORM: score += 400; break;
-            case PL_FMT_FLOAT: score += 300; break;
+            case PL_FMT_UNORM: score += 3; break;
+            case PL_FMT_SNORM: score += 2; break;
+            case PL_FMT_FLOAT: score += 1; break;
             case PL_FMT_TYPE_COUNT: pl_unreachable();
             };
 
@@ -271,12 +285,21 @@ static void set_hdr_metadata(struct priv *p, const struct pl_hdr_metadata *metad
     if (!vk->SetHdrMetadataEXT)
         return;
 
+    // Whitelist only values that we support signalling metadata for
+    struct pl_hdr_metadata fix = {
+        .prim     = metadata->prim,
+        .min_luma = metadata->min_luma,
+        .max_luma = metadata->max_luma,
+        .max_cll  = metadata->max_cll,
+        .max_fall = metadata->max_fall,
+    };
+
     // Ignore no-op changes
-    if (pl_hdr_metadata_equal(metadata, &p->hdr_metadata))
+    if (pl_hdr_metadata_equal(&fix, &p->hdr_metadata))
         return;
 
     // Remember the metadata so we can re-apply it after swapchain recreation
-    p->hdr_metadata = *metadata;
+    p->hdr_metadata = fix;
 
     // Ignore HDR metadata requests for SDR swapchains
     if (!pl_color_transfer_is_hdr(p->color_space.transfer))
@@ -287,18 +310,29 @@ static void set_hdr_metadata(struct priv *p, const struct pl_hdr_metadata *metad
 
     vk->SetHdrMetadataEXT(vk->dev, 1, &p->swapchain, &(VkHdrMetadataEXT) {
         .sType = VK_STRUCTURE_TYPE_HDR_METADATA_EXT,
-        .displayPrimaryRed   = { metadata->prim.red.x,   metadata->prim.red.y },
-        .displayPrimaryGreen = { metadata->prim.green.x, metadata->prim.green.y },
-        .displayPrimaryBlue  = { metadata->prim.blue.x,  metadata->prim.blue.y },
-        .whitePoint = { metadata->prim.white.x, metadata->prim.white.y },
-        .maxLuminance = metadata->max_luma,
-        .minLuminance = metadata->min_luma,
-        .maxContentLightLevel = metadata->max_cll,
-        .maxFrameAverageLightLevel = metadata->max_fall,
+        .displayPrimaryRed   = { fix.prim.red.x,   fix.prim.red.y },
+        .displayPrimaryGreen = { fix.prim.green.x, fix.prim.green.y },
+        .displayPrimaryBlue  = { fix.prim.blue.x,  fix.prim.blue.y },
+        .whitePoint = { fix.prim.white.x, fix.prim.white.y },
+        .maxLuminance = fix.max_luma,
+        .minLuminance = fix.min_luma,
+        .maxContentLightLevel = fix.max_cll,
+        .maxFrameAverageLightLevel = fix.max_fall,
     });
 
     // Keep track of applied HDR colorimetry metadata
-    p->color_space.hdr = *metadata;
+    p->color_space.hdr = p->hdr_metadata;
+}
+
+// Heuristic to determine if we have actual support for VK_EXT_swapchain_colorspace
+static bool check_color_management_support(const struct priv *p)
+{
+    for (int i = 0; i < p->formats.num; i++) {
+        if (p->formats.elem[i].colorSpace != VK_COLOR_SPACE_SRGB_NONLINEAR_KHR && p->formats.elem[i].colorSpace != VK_COLOR_SPACE_PASS_THROUGH_EXT) {
+            return true;
+        }
+    }
+    return false;
 }
 
 pl_swapchain pl_vulkan_create_swapchain(pl_vulkan plvk,
@@ -313,12 +347,12 @@ pl_swapchain pl_vulkan_create_swapchain(pl_vulkan plvk,
     }
 
     struct pl_swapchain_t *sw = pl_zalloc_obj(NULL, sw, struct priv);
-    sw->impl = &vulkan_swapchain;
     sw->log = vk->log;
     sw->gpu = gpu;
 
     struct priv *p = PL_PRIV(sw);
     pl_mutex_init(&p->lock);
+    p->impl = vulkan_swapchain;
     p->params = *params;
     p->vk = vk;
     p->surf = params->surface;
@@ -330,12 +364,19 @@ pl_swapchain pl_vulkan_create_swapchain(pl_vulkan plvk,
         .sType = VK_STRUCTURE_TYPE_SWAPCHAIN_CREATE_INFO_KHR,
         .surface = p->surf,
         .imageArrayLayers = 1, // non-stereoscopic
-        .imageSharingMode = VK_SHARING_MODE_EXCLUSIVE,
+        .imageSharingMode = vk->pools.num > 1 ? VK_SHARING_MODE_CONCURRENT
+                                              : VK_SHARING_MODE_EXCLUSIVE,
+        .queueFamilyIndexCount = vk->pools.num,
+        .pQueueFamilyIndices = p->queue_families,
         .minImageCount = p->swapchain_depth + 1, // +1 for the FB
         .presentMode = params->present_mode,
         .clipped = true,
     };
 
+    pl_assert(vk->pools.num <= PL_ARRAY_SIZE(p->queue_families));
+    for (int i = 0; i < vk->pools.num; i++)
+        p->queue_families[i] = vk->pools.elem[i]->qf;
+
     // These fields will be updated by `vk_sw_recreate`
     p->color_space = pl_color_space_unknown;
     p->color_repr = (struct pl_color_repr) {
@@ -369,6 +410,8 @@ pl_swapchain pl_vulkan_create_swapchain(pl_vulkan plvk,
     VK(vk->GetPhysicalDeviceSurfaceFormatsKHR(vk->physd, p->surf, &num_formats, p->formats.elem));
     p->formats.num = num_formats;
 
+    p->have_color_management_support = check_color_management_support(p);
+
     PL_INFO(gpu, "Available surface configurations:");
     for (int i = 0; i < p->formats.num; i++) {
         PL_INFO(gpu, "    %d: %-40s %s", i,
@@ -377,7 +420,7 @@ pl_swapchain pl_vulkan_create_swapchain(pl_vulkan plvk,
     }
 
     // Ensure there exists at least some valid renderable surface format
-    struct pl_color_space hint = {0};
+    struct pl_color_space hint = pl_color_space_srgb;
     if (!pick_surf_format(sw, &hint))
         goto error;
 
@@ -408,10 +451,10 @@ static void vk_sw_destroy(pl_swapchain sw)
 
     for (int i = 0; i < p->images.num; i++)
         pl_tex_destroy(gpu, &p->images.elem[i]);
-    for (int i = 0; i < p->sems.num; i++) {
-        vk->DestroySemaphore(vk->dev, p->sems.elem[i].in, PL_VK_ALLOC);
-        vk->DestroySemaphore(vk->dev, p->sems.elem[i].out, PL_VK_ALLOC);
-    }
+    for (int i = 0; i < p->sems_in.num; i++)
+        vk->DestroySemaphore(vk->dev, p->sems_in.elem[i], PL_VK_ALLOC);
+    for (int i = 0; i < p->sems_out.num; i++)
+        vk->DestroySemaphore(vk->dev, p->sems_out.elem[i], PL_VK_ALLOC);
 
     vk->DestroySwapchainKHR(vk->dev, p->swapchain, PL_VK_ALLOC);
     pl_mutex_destroy(&p->lock);
@@ -557,6 +600,8 @@ static void destroy_swapchain(struct vk_ctx *vk, void *swapchain)
     vk->DestroySwapchainKHR(vk->dev, vk_unwrap_handle(swapchain), PL_VK_ALLOC);
 }
 
+VK_CB_FUNC_DEF(destroy_swapchain);
+
 static bool vk_sw_recreate(pl_swapchain sw, int w, int h)
 {
     pl_gpu gpu = sw->gpu;
@@ -565,11 +610,12 @@ static bool vk_sw_recreate(pl_swapchain sw, int w, int h)
 
     VkImage *vkimages = NULL;
     uint32_t num_images = 0;
+    char name[32];
 
-    VkSwapchainCreateInfoKHR sinfo = p->protoInfo;
-    if (!update_swapchain_info(p, &sinfo, w, h))
+    if (!update_swapchain_info(p, &p->protoInfo, w, h))
         return false;
 
+    VkSwapchainCreateInfoKHR sinfo = p->protoInfo;
 #ifdef VK_EXT_full_screen_exclusive
     // Explicitly disallow full screen exclusive mode if possible
     static const VkSurfaceFullScreenExclusiveInfoEXT fsinfo = {
@@ -589,13 +635,20 @@ static bool vk_sw_recreate(pl_swapchain sw, int w, int h)
              sinfo.imageExtent.width,
              sinfo.imageExtent.height);
 
+#ifdef PL_HAVE_UNIX
+    if (vk->props.vendorID == VK_VENDOR_ID_NVIDIA) {
+        vk->DeviceWaitIdle(vk->dev);
+        vk_wait_idle(vk);
+    }
+#endif
+
     // Calling `vkCreateSwapchainKHR` puts sinfo.oldSwapchain into a retired
     // state whether the call succeeds or not, so we always need to garbage
     // collect it afterwards - asynchronously as it may still be in use
     sinfo.oldSwapchain = p->swapchain;
     p->swapchain = VK_NULL_HANDLE;
     VkResult res = vk->CreateSwapchainKHR(vk->dev, &sinfo, PL_VK_ALLOC, &p->swapchain);
-    vk_dev_callback(vk, (vk_cb) destroy_swapchain, vk, vk_wrap_handle(sinfo.oldSwapchain));
+    vk_dev_callback(vk, VK_CB_FUNC(destroy_swapchain), vk, vk_wrap_handle(sinfo.oldSwapchain));
     PL_VK_ASSERT(res, "vk->CreateSwapchainKHR(...)");
 
     // Get the new swapchain images
@@ -603,24 +656,25 @@ static bool vk_sw_recreate(pl_swapchain sw, int w, int h)
     vkimages = pl_calloc_ptr(NULL, num_images, vkimages);
     VK(vk->GetSwapchainImagesKHR(vk->dev, p->swapchain, &num_images, vkimages));
 
-    for (int i = 0; i < num_images; i++)
-        PL_VK_NAME(IMAGE, vkimages[i], "swapchain");
+    static const VkSemaphoreCreateInfo seminfo = {
+        .sType = VK_STRUCTURE_TYPE_SEMAPHORE_CREATE_INFO,
+    };
 
     // If needed, allocate some more semaphores
-    while (num_images > p->sems.num) {
-        VkSemaphore sem_in = VK_NULL_HANDLE, sem_out = VK_NULL_HANDLE;
-        static const VkSemaphoreCreateInfo seminfo = {
-            .sType = VK_STRUCTURE_TYPE_SEMAPHORE_CREATE_INFO,
-        };
-        VK(vk->CreateSemaphore(vk->dev, &seminfo, PL_VK_ALLOC, &sem_in));
-        VK(vk->CreateSemaphore(vk->dev, &seminfo, PL_VK_ALLOC, &sem_out));
-        PL_VK_NAME(SEMAPHORE, sem_in, "swapchain in");
-        PL_VK_NAME(SEMAPHORE, sem_out, "swapchain out");
+    while (p->sems_in.num < num_images + 1) { // +1 in case all images are queued
+        VkSemaphore sem = VK_NULL_HANDLE;
+        VK(vk->CreateSemaphore(vk->dev, &seminfo, PL_VK_ALLOC, &sem));
+        snprintf(name, sizeof(name), "swapchain in #%d", p->sems_in.num);
+        PL_VK_NAME(SEMAPHORE, sem, name);
+        PL_ARRAY_APPEND(sw, p->sems_in,  sem);
+    }
 
-        PL_ARRAY_APPEND(sw, p->sems, (struct sem_pair) {
-            .in = sem_in,
-            .out = sem_out,
-        });
+    while (p->sems_out.num < num_images) {
+        VkSemaphore sem = VK_NULL_HANDLE;
+        VK(vk->CreateSemaphore(vk->dev, &seminfo, PL_VK_ALLOC, &sem));
+        snprintf(name, sizeof(name), "swapchain out #%d", p->sems_out.num);
+        PL_VK_NAME(SEMAPHORE, sem, name);
+        PL_ARRAY_APPEND(sw, p->sems_out,  sem);
     }
 
     // Recreate the pl_tex wrappers
@@ -630,12 +684,14 @@ static bool vk_sw_recreate(pl_swapchain sw, int w, int h)
 
     for (int i = 0; i < num_images; i++) {
         const VkExtent2D *ext = &sinfo.imageExtent;
+        snprintf(name, sizeof(name), "swapchain #%d", i);
         pl_tex tex = pl_vulkan_wrap(gpu, pl_vulkan_wrap_params(
             .image = vkimages[i],
             .width = ext->width,
             .height = ext->height,
             .format = sinfo.imageFormat,
             .usage = sinfo.imageUsage,
+            .debug_tag = name,
         ));
         if (!tex)
             goto error;
@@ -657,7 +713,29 @@ static bool vk_sw_recreate(pl_swapchain sw, int w, int h)
     p->color_repr.bits.color_depth = bits;
 
     // Note: `p->color_space.hdr` is (re-)applied by `set_hdr_metadata`
-    map_color_space(sinfo.imageColorSpace, &p->color_space);
+    if (p->have_color_management_support) {
+        map_color_space(sinfo.imageColorSpace, &p->color_space);
+    } else {
+        // Compositor doesn't have color management support
+        switch (sinfo.imageColorSpace) {
+        case VK_COLOR_SPACE_SRGB_NONLINEAR_KHR:
+            p->color_space = pl_color_space_monitor;
+            break;
+        case VK_COLOR_SPACE_PASS_THROUGH_EXT:
+            p->color_space = pl_color_space_unknown;
+            break;
+        default:
+            // This should never happen
+            pl_unreachable();
+        }
+    }
+
+    // To convert to XYZ color system for VK_COLOR_SPACE_DCI_P3_NONLINEAR_EXT
+    if (p->color_space.transfer == PL_COLOR_TRC_ST428) {
+        p->color_repr.sys = PL_COLOR_SYSTEM_XYZ;
+    } else {
+        p->color_repr.sys = PL_COLOR_SYSTEM_RGB;
+    }
 
     // Forcibly re-apply HDR metadata, bypassing the no-op check
     struct pl_hdr_metadata metadata = p->hdr_metadata;
@@ -692,7 +770,8 @@ static bool vk_sw_start_frame(pl_swapchain sw,
         return false;
     }
 
-    VkSemaphore sem_in = p->sems.elem[p->idx_sems].in;
+    VkSemaphore sem_in = p->sems_in.elem[p->idx_sems_in];
+    p->idx_sems_in = (p->idx_sems_in + 1) % p->sems_in.num;
     PL_TRACE(vk, "vkAcquireNextImageKHR signals 0x%"PRIx64, (uint64_t) sem_in);
 
     for (int attempts = 0; attempts < 2; attempts++) {
@@ -748,6 +827,8 @@ static void present_cb(struct priv *p, void *arg)
     (void) pl_rc_deref(&p->frames_in_flight);
 }
 
+VK_CB_FUNC_DEF(present_cb);
+
 static bool vk_sw_submit_frame(pl_swapchain sw)
 {
     pl_gpu gpu = sw->gpu;
@@ -756,8 +837,7 @@ static bool vk_sw_submit_frame(pl_swapchain sw)
     pl_assert(p->last_imgidx >= 0);
     pl_assert(p->swapchain);
     uint32_t idx = p->last_imgidx;
-    VkSemaphore sem_out = p->sems.elem[p->idx_sems++].out;
-    p->idx_sems %= p->sems.num;
+    VkSemaphore sem_out = p->sems_out.elem[idx];
     p->last_imgidx = -1;
 
     bool held = pl_vulkan_hold_ex(gpu, pl_vulkan_hold_params(
@@ -780,8 +860,8 @@ static bool vk_sw_submit_frame(pl_swapchain sw)
     }
 
     pl_rc_ref(&p->frames_in_flight);
-    vk_cmd_callback(cmd, (vk_cb) present_cb, p, NULL);
-    if (!vk_cmd_submit(vk, &cmd)) {
+    vk_cmd_callback(cmd, VK_CB_FUNC(present_cb), p, NULL);
+    if (!vk_cmd_submit(&cmd)) {
         pl_mutex_unlock(&p->lock);
         return false;
     }
@@ -879,7 +959,7 @@ bool pl_vulkan_swapchain_suboptimal(pl_swapchain sw)
     return p->suboptimal;
 }
 
-static struct pl_sw_fns vulkan_swapchain = {
+static const struct pl_sw_fns vulkan_swapchain = {
     .destroy            = vk_sw_destroy,
     .latency            = vk_sw_latency,
     .resize             = vk_sw_resize,
